%CUDABEAMPROP A CUDA-MATLAB implementation of a simple 2D FFT based
% paraxial beam propagation method
%
% dE/dz = 1/(2ibeta)( (d/dx^2 + d/dy^2) - (beta^2 - kzero^2*n^2) )E
%
% Author: Patrick Kano, Paul Lundquist, Eric Nelson-Melby
% Applied Energetics, Inc.
% Modification Dates [M/D/Y]: 10/21/10 - Initial script work
% Modification Dates [M/D/Y]: 11/04/10 - Jacketize
% References: 
% www.accelereyes.com
% A. Weideman, Linear Dispersive Wave Equations, 
% http://dip.sun.ac.za/~weideman/research/waves.html
% Okamoto, K.  Fundamentals of optical waveguides, Academic Press, 2000.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [RunTimeOut,Efield,VecposX,VecposY] = CUDAparaxbpm(CUDAswitch, PLOTswitch, BetaSwitch, Nx, Ny, LengthX, LengthY, LengthZ, deltaZ, Lambda)

if(nargin==0)
 CUDAswitch = 1; %0-CPU fft2, 1-GPU fft2
 PLOTswitch = 1; %0-No plots, 1-With plots
 BetaSwitch = 1; %0-Constant beta, 1-Update beta

 LengthX = 4*10^(-3);  %m
 LengthY = 4*10^(-3);  %m
 LengthZ = 5*10^(-3);     %m
 deltaZ = LengthZ/1000; %m

 Lambda = 1000*10^(-6); %m

 Nx = 2^6; %Must be 2^N
 Ny = 2^6; %Must be 2^N
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DO NOT MODIFY BELOW THIS LINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tic;

%Check for Jacket installation
if(~exist('ginfo'))
 disp('Jacket is not available.');
 disp('Defaulting to a computation with standard MATLAB.');
 CUDAswitch = 0;
else
 % Jacket supports all CUDA GPUs so no need to check for CUDA version.
end

kzerolambda = 2*pi/Lambda;

%%Setup the grid and the wavenumbers
nXidx = -Nx:Nx-1;
deltaX = LengthX/Nx;
VecposX = nXidx*deltaX; %[-Lx,Lx]
VeckX = nXidx*pi/LengthX;

nYidx = -Ny:Ny-1;
deltaY = LengthY/Ny;
VecposY = nYidx*deltaY; %[-Ly,Ly]
VeckY = nYidx*pi/LengthY;

Nz = floor(LengthZ/deltaZ);
VecposZ = linspace(0,LengthZ,Nz);

if CUDAswitch==1,
  [MeshX, MeshY] = meshgrid(gdouble(VecposX),gdouble(VecposY));
  [MeshkX, MeshkY] = meshgrid(gdouble(VeckX),gdouble(VeckY));
else
  [MeshX, MeshY] = meshgrid(VecposX,VecposY);
  [MeshkX, MeshkY] = meshgrid(VeckX,VeckY);
end

MeshSquarekxy = MeshkX.*MeshkX + MeshkY.*MeshkY;
MeshSquarekxy = fftshift(MeshSquarekxy);

%%Intiailize the E-field to ones
Efield = FunEfieldinit(MeshX, MeshY, Nx, Ny, LengthX, LengthY);

nMatrix = FunRefractIdx(MeshX, MeshY, Nx, Ny, 0);

Beta = kzerolambda*sqrt( sum(sum(abs(nMatrix.*Efield).^2)) / sum(sum(abs(Efield).^2))); %Ignore spatial E-gradients
BetaStoreVec = kzerolambda*ones(Nz,1);

FreeSpaceProp = exp(1i*deltaZ*MeshSquarekxy/(2*Beta));
%if CUDAswitch==1, gFreeSpaceProp = gdouble(FreeSpaceProp);  end %Put on the GPU

%%Z-updates
for zidx=1:Nz
 PosZ = VecposZ(zidx);
 nMatrix = FunRefractIdx(MeshX, MeshY, Nx, Ny, PosZ);

 if BetaSwitch==1 %Updated beta methods
  Beta = kzerolambda*sqrt( sum(sum(abs(nMatrix.*Efield).^2)) / sum(sum(abs(Efield).^2))); %Ignore spatial E-gradients
  BetaStoreVec(zidx) = Beta; 
   
  FreeSpaceProp = exp(1i*deltaZ*MeshSquarekxy/(2*Beta));
  %if CUDAswitch==1, gFreeSpaceProp = gdouble(FreeSpaceProp);  end %Put on the GPU   
 end
  
 PhaseScreen = exp((0.5*deltaZ/(-2i*Beta))*(Beta*Beta*ones(2*Nx,2*Ny) - kzerolambda*kzerolambda*nMatrix.^2));
 
 %Strang-splitting
 Efield = PhaseScreen.*Efield;
 if CUDAswitch==0, Efield = ifft2( FreeSpaceProp.*fft2(Efield) );
 else Efield = ifft2( FreeSpaceProp.*fft2(Efield) );  
 end
 Efield = PhaseScreen.*Efield;
 
 if PLOTswitch==1, FunPlotSim(zidx, PosZ, MeshX, MeshY, nMatrix, Efield, MeshkX, MeshkY, MeshSquarekxy); end   
end %z for loop

Efield = double( Efield );

if(PLOTswitch==1)
 figure(1004);
 plot(VecposZ,BetaStoreVec,'-ko');
 xlabel('z [m]');
 ylabel('\beta');
end 
 
RunTimeOut=toc;
end %function definition
%end of file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%